/****************************************************************************
    
    This file is part of The Newtonator soft synth.

    Copyright (C) 2010-2011 Michael Bechard <mbechard@users.sourceforge.net>

    The Newtonator is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The Newtonator is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with The Newtonator.  If not, see <http://www.gnu.org/licenses/>.

****************************************************************************/

# include "Newtonator.h"
#include <cmath>
#include <stdlib.h>

using namespace stk;

Newtonator::Newtonator(bool useNoteEnvelope, unsigned int channels)
: MAX_VAL(1.0), MIN_VAL(-1.0)
{
    _pPrevPos = NULL;
    _pPrevFloor = NULL;
    _pPrevVeloc = NULL;
    _gravDir = NULL;
    _floorSource = NULL;
    _gravModSource = NULL;
    _ampEnv = NULL;
    _numChannels = 0;
    _floorAttraction = 0.0;
    _floorScale = 0.5;
    _gravModDepth = 0.0;
    _gravity = 0.002441406; //works out to ~80 in csound...
    _gravScale = 0.585806438; //works out to .5 in csound...
    _noteAmp = 0.0;
    //_dcLevel = 0.0;
    _clip = false;
    _usingNoteEnvelope = useNoteEnvelope;
    initChannels(channels);
}

Newtonator::~Newtonator()
{
    clearChannelMemory();
}

void Newtonator::clearChannelMemory()
{
    if (_pPrevPos != NULL) delete [] _pPrevPos;
    if (_pPrevFloor != NULL) delete [] _pPrevFloor;
    if (_pPrevVeloc != NULL) delete [] _pPrevVeloc;
    if (_gravDir != NULL) delete [] _gravDir;
    if (_floorSource != NULL) delete [] _floorSource;
    if (_gravModSource != NULL) delete [] _gravModSource;
    if (_ampEnv != NULL) delete [] _ampEnv;
    _pPrevPos = NULL;
    _pPrevFloor = NULL;
    _pPrevVeloc = NULL;
    _gravDir = NULL;
    _floorSource = NULL;
    _gravModSource = NULL;
    _ampEnv = NULL;
    _numChannels = 0;
}

void Newtonator::initChannels(unsigned int numChannels)
{
    //If the number of channels has changed, reset all channel
    //memory variables.
    if (numChannels != _numChannels && numChannels > 0)
    {
        clearChannelMemory();
        _pPrevPos = new StkFloat[numChannels];
        _pPrevFloor = new StkFloat[numChannels];
        _pPrevVeloc = new StkFloat[numChannels];
        _gravDir = new int[numChannels];
        _floorSource = new VariSource[numChannels];
        _gravModSource = new VariSource[numChannels];
        _ampEnv = new ADSR[numChannels];
        _numChannels = numChannels;

        for (unsigned int i = 0; i < _numChannels; i++)
        {
            _pPrevPos[i] = rand() / RAND_MAX;
            _pPrevFloor[i] = 0.0;
            _pPrevVeloc[i] = 0.0;
            _gravDir[i] = (i % 2 == 0 ? 1 : -1);
            _floorSource[i].setUsingEnvelope(false);
            if (!_usingNoteEnvelope) _gravModSource[i].setUsingEnvelope(false);
        }
    }
}

void Newtonator::keyOn(stk::StkFloat velocity)
{
    for (unsigned int chan = 0; chan < _numChannels; chan++)
    {
        _ampEnv[chan].keyOn();
        _noteAmp = velocity;
        _gravModSource[chan].setAmplitude(velocity);
        _gravModSource[chan].getEnv().keyOn();
    }
}

void Newtonator::keyOff(stk::StkFloat velocity)
{
    for (unsigned int chan = 0; chan < _numChannels; chan++)
    {
        _ampEnv[chan].keyOff();
        _gravModSource[chan].getEnv().keyOff();
    }
}

bool Newtonator::isPlaying()
{
    if (!_usingNoteEnvelope) return true;

    for (unsigned int chan = 0; chan < _numChannels; chan++)
    {
        if (_ampEnv[chan].getState() != ADSR::DONE) return true;
    }

    return false;
}

StkFloat Newtonator::newtTick(unsigned int channel, unsigned int frameIdx)
{
    StkFloat retSample;

    if (_numChannels == 0) return 0.0;

    //Initiate channel "memory" of floor and positions
    StkFloat& prevPos = _pPrevPos[channel];
    StkFloat& prevFloor = _pPrevFloor[channel];
    StkFloat& prevVeloc = _pPrevVeloc[channel];
    int& gravDir = _gravDir[channel];
    StkFloat floor = _floorScale * _floorSource[channel].tick(frameIdx);
    StkFloat gravMod = _gravModDepth * _gravModSource[channel].tick(frameIdx);
    StkFloat gravity = _gravity * (1 + gravMod);
    StkFloat tempPos;
    StkFloat gravScale;

    //Main computation
    if (std::abs(prevPos - prevFloor) < (_floorAttraction * MAX_VAL) &&
        prevVeloc < floor - prevFloor &&
        prevVeloc > _floorAttraction * gravDir * gravity * Stk::sampleRate())
    {
        //vertical velocity isn't enough to overcome slope in floor,
        //calculate new velocity
        tempPos = floor;
        prevVeloc = floor - prevFloor;
    }
    else
    {
        //calculate current y-position (t is always 1)
        StkFloat temp = (prevPos - prevFloor) / (MAX_VAL * _gravScale);
        gravScale = sqrt(std::abs(temp));
        tempPos = prevPos + prevVeloc + (gravDir * gravity * gravScale / 2);
        prevVeloc = prevVeloc + (gravDir * gravity);
    }

    //normalize values with floor/ceiling
    if ((gravDir < 0 && tempPos < floor) || (gravDir > 0 && tempPos >= floor))
    {
        gravDir *= -1;
    }
    if (!_clip)
    {
        retSample = tempPos;
    }
    else if (tempPos > 0)
    {
        retSample = std::min(MAX_VAL, tempPos);
    }
    else if (tempPos < 0)
    {
        retSample = std::max(MIN_VAL, tempPos);
    }

    //remember channel values (prevVeloc was set earlier)
    prevPos = retSample;
    prevFloor = floor;

    //squash and trim to underlying floor waveform
    retSample = retSample * (1 - _floorScale);
    StkFloat finalCeil = floor + (1 - _floorScale);
    StkFloat finalFloor = floor - (1 - _floorScale);
    if (retSample > finalCeil)
        retSample = finalCeil;
    else if (retSample < finalFloor)
        retSample = finalFloor;

    //apply note envelope, if applicable
    if (_usingNoteEnvelope)
    {
        StkFloat env = _ampEnv[channel].tick();
        retSample = retSample * _noteAmp * env;
    }

    return retSample;
}
